<?php

namespace IMATHUZH\Qfq\Core\Renderer;

use IMATHUZH\Qfq\Core\Form\Form;
use IMATHUZH\Qfq\Core\Helper\OnArray;
use IMATHUZH\Qfq\Core\Helper\Support;
use IMATHUZH\Qfq\Core\Renderer\Element\Bootstrap3\Bootstrap3ButtonRenderer;
use IMATHUZH\Qfq\Core\Renderer\Element\Bootstrap3\Bootstrap3SpinnerRenderer;
use IMATHUZH\Qfq\Core\Renderer\FormElement\Bootstrap3\Bootstrap3CheckboxRenderer;
use IMATHUZH\Qfq\Core\Renderer\FormElement\Bootstrap3\Bootstrap3DatetimeRenderer;
use IMATHUZH\Qfq\Core\Renderer\FormElement\Bootstrap3\Bootstrap3EditorRenderer;
use IMATHUZH\Qfq\Core\Renderer\FormElement\Bootstrap3\Bootstrap3FieldsetRenderer;
use IMATHUZH\Qfq\Core\Renderer\FormElement\Bootstrap3\Bootstrap3InputRenderer;
use IMATHUZH\Qfq\Core\Renderer\FormElement\Bootstrap3\Bootstrap3NoteRenderer;
use IMATHUZH\Qfq\Core\Renderer\FormElement\Bootstrap3\Bootstrap3PillRenderer;
use IMATHUZH\Qfq\Core\Renderer\FormElement\Bootstrap3\Bootstrap3RadioButtonRenderer;
use IMATHUZH\Qfq\Core\Renderer\FormElement\Bootstrap3\Bootstrap3SelectRenderer;
use IMATHUZH\Qfq\Core\Renderer\FormElement\Bootstrap3\Bootstrap3SubrecordRenderer;

class Bootstrap3Renderer extends BaseRenderer
{
    public function __construct()
    {
        // If Bootstrap3Renderer does not specify a Renderer for a specific element, then the base version of that renderer will be used!
        parent::__construct();

        //Bootstrap3-specific Renderers
        $this->buttonRenderer = new Bootstrap3ButtonRenderer();
        $this->spinnerRenderer = new Bootstrap3SpinnerRenderer();

        $this->pillRenderer = new Bootstrap3PillRenderer();
        $this->fieldsetRenderer = new Bootstrap3FieldsetRenderer();

        $this->noteRenderer = new Bootstrap3NoteRenderer();
        $this->inputRenderer = new Bootstrap3InputRenderer();
        $this->checkboxRenderer = new Bootstrap3CheckboxRenderer();
        $this->radioRenderer = new Bootstrap3RadioButtonRenderer();
        $this->selectRenderer = new Bootstrap3SelectRenderer();
        $this->subrecordRenderer = new Bootstrap3SubrecordRenderer();
        $this->editorRenderer = new Bootstrap3EditorRenderer();
        $this->datetimeRenderer = new Bootstrap3DatetimeRenderer();
        //$this->uploadRenderer = new SpinnerRenderer();
        //$this->extraRenderer = new SpinnerRenderer();
    }

    /**
     * @param array $formSpec Associative array containing all form specifications
     * @param string $renderMode HTML/JSON
     * @return string Rendered Form Head
     */
    protected function renderFormHead(Form $form, string $renderMode = RENDER_MODE_HTML) : string {
        $formSpec = $form->specFinal;
        $output = '';

        switch ($renderMode){
            case RENDER_MODE_JSON_CLIENT:
            case RENDER_MODE_JSON_DYNAMIC:
            case RENDER_MODE_HTML:
                $containerClassAttribute = Support::doAttribute('class', $formSpec[F_CLASS], true);

                // Show title / frame only if there is a title given.
                $titleHtml = "";
                if (trim($form->specFinal[F_TITLE]) != '') {
                    $classTitle = isset($form->specFinal[F_CLASS_TITLE]) ? $form->specFinal[F_CLASS_TITLE] : "qfq-form-title";
                    $titleHtml = Support::wrapTag('<div class="row"><div class="col-md-12">', Support::wrapTag('<div class="' . $classTitle . '">', $form->specFinal[F_TITLE]));
                }

                // Prepare Form Tag Attributes
                $formTagAttributes = $form->formTagAttributes;
                $formTagAttributes['class'] = 'form-horizontal qfq-form';
                $formTagAttributes['data-toggle'] = 'validator';
                $formTagAttributes = OnArray::toString($formTagAttributes, '=', ' ', "'");

                // Prepare Pills (if any)
                $pillHtml = $this->renderPillNavigation($form);

                //Prepare Content Div Classes
                $contentDivClassesArray = ['tab-content', 'qfq-color-grey-2', $form->specFinal[F_CLASS_BODY]];
                if ($pillHtml == '') {
                    $contentDivClassesArray[] = 'col-md-12';
                    $contentDivClassesArray[] = 'qfq-form-body'; // Make an outline on form body

                    if ($titleHtml == '') {
                        $contentDivClassesArray[] = 'qfq-form-no-title';
                    }
                }
                $contentDivClassAttribute = Support::doAttribute('class', $contentDivClassesArray);

                // Prepare Buttons above Form
                $controlButtonsHtml = $this->renderControlButtonsHtml($form);
                $honeypotHtml = $this->renderHoneypotElements();
                $hiddenSip = $this->renderHiddenSip();
                $hiddenT3Sip = $this->renderHiddenT3Sip();
                $recordHashMd5InputHtml = $this->renderRecordHashMd5Input($form);

                $output .= <<<EOL
                                <div $containerClassAttribute>
                                    <div class="row">
                                        <div class="col-md-12">
                                            $controlButtonsHtml
                                        </div>
                                    </div>
                                    $titleHtml
                                    <div class="row">
                                        $pillHtml
                                    </div>

                                    <form $formTagAttributes>
                                        $honeypotHtml
                                        $hiddenSip
                                        $hiddenT3Sip
                                        $recordHashMd5InputHtml
                                        <div $contentDivClassAttribute>
                    EOL;
        }

        return $output;
    }

    protected function renderFormTail(Form $form, string $renderMode = RENDER_MODE_HTML) : string {

        $output = '';
        $formSpec = $form->specFinal;

        switch ($renderMode) {

            case RENDER_MODE_JSON_CLIENT:
            case RENDER_MODE_JSON_DYNAMIC:
            case RENDER_MODE_HTML:

            $output .= '</div>';  //  <div class="tab-content">

            // Button Save at bottom of form - only if there is a button text given.
            if ($formSpec[F_SUBMIT_BUTTON_TEXT] !== '') {

                // Default setzen:
                //TODO $this->fillWrapLabelInputNote($formSpec[F_BS_LABEL_COLUMNS], $formSpec[F_BS_INPUT_COLUMNS], $formSpec[F_BS_NOTE_COLUMNS]);

                $buttonText = $formSpec[F_SUBMIT_BUTTON_TEXT];

                /* TODO $htmlElement = $this->buildButtonCode('save-button', $buttonText, $formSpec[F_SUBMIT_BUTTON_TOOLTIP],
                    $formSpec[F_SUBMIT_BUTTON_GLYPH_ICON], '', $formSpec[F_BUTTON_ON_CHANGE_CLASS],
                    $formSpec[F_SUBMIT_BUTTON_CLASS]); */

                //New save button style if save button text is given
                $lastDiv = '</div></div>';

                //finish elements and close tag-form before wrapping save button
                $labelClass = $formSpec[FE_BS_LABEL_COLUMNS];
                $inputClass = $formSpec[FE_BS_INPUT_COLUMNS];
                $noteClass = $formSpec[FE_BS_NOTE_COLUMNS];

                $output .= <<<EOF
                            <div class="row">
                                <div class="col-md-12" style="height: 45px; padding-top: 8px;">
                                    <div class="$labelClass qfq-label"></div>
                                    <div class="$inputClass">
                                        <!-- htmlElement here (Save button at bottom)-->
                                    </div>
                                    <div class="$noteClass qfq-note"></div>
                                </div>
                            </div>
                            <div class="clearfix"></div>
                EOF;
            }

            $output .= '</form></div>';  //<div class="container-fluid"> === main <div class=...> around everything

        }

        return $output;

    }

    /**
     * Loops through all formElements of the given form and render the navigation for all pills found.
     * @param Form $form
     * @return string
     */
    private function renderPillNavigation(Form $form) : string{

        $pillButton = '';
        $pillDropdown = '';
        $htmlDropdown = '';
        $ii = 0;
        $isFirstPill = true;

        for ($i = 0; $i < count($form->formElements); $i++) {
            if($form->formElements[$i]->attributes[FE_TYPE] !== FE_TYPE_PILL) continue;


            // TODO: THE FOLLOWING
            //if ($mode != FORM_LOAD && $formElement[FE_DYNAMIC_UPDATE] !== 'yes') {
            //    continue; // During save/update: Process only FE dynamic_update=yes
            //}

            $htmlIdLi = $form->formElements[$i]->attributes[FE_HTML_ID] . HTML_ID_EXTENSION_PILL_LI;
            $htmlIdLiA = $form->formElements[$i]->attributes[FE_HTML_ID] . HTML_ID_EXTENSION_PILL_LI_A;

            $ii++;

            $attributeLiA = 'data-toggle="tab" ';

            switch ($form->formElements[$i]->attributes[FE_MODE]) {
                case FE_MODE_SHOW:
                case FE_MODE_REQUIRED:
                    $attributeLi = '';
                    $pillHref = $form->formElements[$i]->htmlAttributes["href"];
                    //$json[$htmlFormElementName][API_ELEMENT_UPDATE][$htmlIdLi][API_ELEMENT_ATTRIBUTE][HTML_ATTR_CLASS] = '';
                    //$json[$htmlFormElementName][API_ELEMENT_UPDATE][$htmlIdLiA][API_ELEMENT_ATTRIBUTE][HTML_ATTR_CLASS] = '';
                    break;

                case FE_MODE_READONLY:
                    $pillHref = '#';
                    $attributeLi = Support::doAttribute('class', 'disabled');
                    //$json[$htmlFormElementName][API_ELEMENT_UPDATE][$htmlIdLi][API_ELEMENT_ATTRIBUTE][HTML_ATTR_CLASS] = 'disabled';
                    $attributeLiA .= Support::doAttribute('class', 'noclick');
                    //$json[$htmlFormElementName][API_ELEMENT_UPDATE][$htmlIdLiA][API_ELEMENT_ATTRIBUTE][HTML_ATTR_CLASS] = 'noclick';
                    break;

                case FE_MODE_HIDDEN:
                    $attributeLi = Support::doAttribute('class', 'hidden');
                    //$json[$htmlFormElementName][API_ELEMENT_UPDATE][$htmlIdLi][API_ELEMENT_ATTRIBUTE][HTML_ATTR_CLASS] = 'hidden';
                    break;

                default:
                    throw new \UserFormException("Unknown Mode: " . $form->formElements[$i]->attributes[FE_MODE], ERROR_UNKNOWN_MODE);
            }

            $attributeLi .= Support::doAttribute(HTML_ATTR_ID, $htmlIdLi);
            $attributeLi .= Support::doAttribute('title', $form->formElements[$i]->attributes[FE_TOOLTIP]);
            $attributeLiA .= Support::doAttribute(HTML_ATTR_ID, $htmlIdLiA);
            $a = Support::wrapTag("<a $attributeLiA" . Support::doAttribute('href', $pillHref) . ">", $form->formElements[$i]->attributes[FE_LABEL]);
            //$json[$htmlFormElementName][API_ELEMENT_UPDATE][$htmlIdLiA][API_ELEMENT_CONTENT] = $formElement[FE_LABEL];

            if ($isFirstPill && $form->formElements[$i]->attributes[FE_MODE] != FE_MODE_HIDDEN) {
                $attributeLi .= 'class="active" ';
                $isFirstPill = false;
            }

            if ($ii <= $form->specFinal[F_MAX_VISIBLE_PILL]) {
                $pillButton .= '<li role="presentation"' . $attributeLi . ">" . $a . "</li>";
            } else {
                $pillDropdown .= '<li ' . $attributeLi . '>' . $a . "</li>";
            }
        }


        // Pill Dropdown necessary?
        if (isset($form->specFinal[F_MAX_VISIBLE_PILL]) && $ii > $form->specFinal[F_MAX_VISIBLE_PILL]) {
            $htmlDropdown = Support::wrapTag('<ul class="dropdown-menu qfq-form-pill ' . $form->specFinal[F_CLASS_PILL] . '">', $pillDropdown, true);
            $htmlDropdown = '<a class="dropdown-toggle" data-toggle="dropdown" href="$pillHref" role="button">more <span class="caret"></span></a>' . $htmlDropdown;
            $htmlDropdown = Support::wrapTag('<li role="presentation" class="dropdown">', $htmlDropdown, false);
        }

        // Check for active remember last pill
        $flagRememberLastPill = ($form->specFinal[F_REMEMBER_LAST_PILL] == '1') ? 'true' : 'false';

        $tabId = 'qfqTabs-' . $this->store->getVar(TYPO3_TT_CONTENT_UID, STORE_TYPO3) . '-'
                            . $form->specFinal[F_ID] . '-' . $this->store->getVar(CLIENT_RECORD_ID, STORE_TYPO3 . STORE_SIP . STORE_RECORD . STORE_ZERO);
        $htmlDropdown = Support::wrapTag('<ul id="' . $tabId . '" data-active-last-pill="' . $flagRememberLastPill . '" class="nav nav-pills qfq-form-pill ' . $form->specFinal[F_CLASS_PILL] . '" role="tablist">', $pillButton . $htmlDropdown);
        $htmlDropdown = Support::wrapTag('<div class="col-md-12">', $htmlDropdown);

        return $htmlDropdown;
    }




    /**
     * Build Buttons panel on top right corner of form.
     *
     * @return string
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     * @throws \UserReportException
     */
    private function renderControlButtonsHtml(Form $form) : string {
        // Debug Buttons (only if logged into T3 Backend)
        $debugButtons = '';
        if ($form->showDebugInfoFlag) {
            $debugButtons = $this->renderViewFormButton($form)
                            . $this->renderEditFormElementButton()
                            . $this->renderCopyFormButton($form)
                            . $this->renderEditFormButton($form);
        }

        $buttonSave = $this->renderSaveButton($form);
        $buttonClose = $this->renderCloseButton($form);
        $buttonDelete = $this->renderDeleteButton($form);
        $buttonNew = $this->renderNewButton($form);

        // Arrangement: Edit Form / Save / Close / Delete / New
        $html = Support::wrapTag('<div class="btn-group" role="group">', $debugButtons, true);
        $html .= Support::wrapTag('<div class="btn-group" role="group">', $buttonSave . $buttonClose, true);
        $html .= Support::wrapTag('<div class="btn-group" role="group">', $buttonDelete, true);
        $html .= Support::wrapTag('<div class="btn-group" role="group">', $buttonNew, true);

        return Support::wrapTag('<div class="btn-toolbar pull-right" role="toolbar">', $html);
    }

    /**
     * Renders the Form Save button for given form
     */
    private function renderSaveButton(Form $form) : string{
        if (Support::findInSet(FORM_BUTTON_SAVE, $form->specFinal[F_SHOW_BUTTON]) && $form->specFinal[F_SUBMIT_BUTTON_TEXT] === '') {
            $toolTip = $form->specFinal[F_SAVE_BUTTON_TOOLTIP];

            if ($toolTip == 'Save') {
                if ($form->recordId == 0) {
                    $toolTip .= PHP_EOL . 'Create new record';
                } else {
                    $toolTip .= PHP_EOL . 'Record id: ' . $form->recordId;
                    $toolTip .= PHP_EOL . 'Created: ' . $this->store->getVar('created', STORE_RECORD . STORE_EMPTY);;
                    $toolTip .= PHP_EOL . 'Modified: ' . $this->store->getVar('modified', STORE_RECORD . STORE_EMPTY);;
                }
            }
            // In debugMode every button link should show the information behind the SIP.
            if ($form->showDebugInfoFlag) {
                $toolTip .= PHP_EOL . "Table: " . $form->specFinal[F_TABLE_NAME];
            }

            $buttonSpec = array();
            $buttonSpec['buttonHtmlId'] = 'save-button-' . $form->formTagAttributes[F_ID];
            $buttonSpec['tooltip'] = $toolTip;
            $buttonSpec['icon'] = $form->specFinal[F_SAVE_BUTTON_GLYPH_ICON];
            $buttonSpec['class'] = $form->specFinal[F_SAVE_BUTTON_CLASS];
            $buttonSpec['onChangeClass'] = $form->specFinal[F_BUTTON_ON_CHANGE_CLASS];
            return $this->buttonRenderer->renderElement($buttonSpec);
        }
        return "";
    }

    /**
     * Renders the Form Close button for given form
     */
    private function renderCloseButton(Form $form) : string{
        if (Support::findInSet(FORM_BUTTON_CLOSE, $form->specFinal[F_SHOW_BUTTON])) {
            $buttonSpec = array();
            $buttonSpec['buttonHtmlId'] = 'close-button-' . $form->formTagAttributes[F_ID];
            $buttonSpec['text'] = $form->specFinal[F_CLOSE_BUTTON_TEXT];
            $buttonSpec['tooltip'] = $form->specFinal[F_CLOSE_BUTTON_TOOLTIP];
            $buttonSpec['icon'] = $form->specFinal[F_CLOSE_BUTTON_GLYPH_ICON];
            $buttonSpec['class'] = $form->specFinal[F_CLOSE_BUTTON_CLASS];
            return $this->buttonRenderer->renderElement($buttonSpec);
        }
        return "";
    }

    /**
     * Renders the Form Delete button for given form
     */
    private function renderDeleteButton(Form $form) : string{
        // Button: Delete
        if (Support::findInSet(FORM_BUTTON_DELETE, $form->specFinal[F_SHOW_BUTTON])) {
            $toolTip = $form->specFinal[F_DELETE_BUTTON_TOOLTIP];

            if ($form->showDebugInfoFlag && $form->recordId > 0) {
                $toolTip .= PHP_EOL . "form = '" . $form->specFinal[F_FINAL_DELETE_FORM] . "'" . PHP_EOL . "r = '" . $form->recordId . "'";
            }
            $disabled = ($form->recordId > 0) ? '' : 'disabled';

            $buttonSpec = array();
            $buttonSpec['buttonHtmlId'] = 'delete-button-' . $form->formTagAttributes[F_ID];
            $buttonSpec['text'] = $form->specFinal[F_DELETE_BUTTON_TEXT];
            $buttonSpec['tooltip'] = $toolTip;
            $buttonSpec['icon'] = $form->specFinal[F_DELETE_BUTTON_GLYPH_ICON];
            $buttonSpec['class'] = $form->specFinal[F_DELETE_BUTTON_CLASS];
            $buttonSpec['disabled'] = $disabled;

            return $this->buttonRenderer->renderElement($buttonSpec);
        }
        return "";
    }

    /**
     * Renders the form new button for given form
     */
    private function renderNewButton(Form $form) : string{
        // Button: New
        if (Support::findInSet(FORM_BUTTON_NEW, $form->specFinal[F_SHOW_BUTTON])) {
            $url = $form->getNewUrl($toolTip);

            $buttonSpec = array();
            $buttonSpec['url'] = $url;
            $buttonSpec['buttonHtmlId'] = 'form-new-button';
            $buttonSpec['text'] = $form->specFinal[F_NEW_BUTTON_TEXT];
            $buttonSpec['tooltip'] = $toolTip;
            $buttonSpec['icon'] = $form->specFinal[F_NEW_BUTTON_GLYPH_ICON];
            $buttonSpec['class'] = $form->specFinal[F_NEW_BUTTON_CLASS];

            return $this->buttonRenderer->renderElement($buttonSpec);
        }
        return "";
    }

    /**
     * Renders the 'Edit Form' button for given form
     * @param Form $form
     * @return void
     */
    private function renderEditFormButton(Form $form) : string{
        $buttonSpec = array();
        $buttonSpec['url'] = $form->getFormEditorUrl();
        $buttonSpec['buttonHtmlId'] = 'form-edit-button';
        $buttonSpec['tooltip'] = "Edit form" . PHP_EOL . PHP_EOL . OnArray::toString($this->store->getStore(STORE_SIP), ' = ', PHP_EOL, "'");
        $buttonSpec['icon'] = GLYPH_ICON_TOOL;
        $buttonSpec['class'] = 'btn btn-default navbar-btn';
        return $this->buttonRenderer->renderElement($buttonSpec);
    }

    /**
     * Creates a link to open current form loaded in FormEditor
     *
     * @return string
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     */
    private function renderViewFormButton(Form $form) : string {
        $buttonSpec = array();


        // START EXISTING CODE
        $formName = false;
        $buttonSpec['url'] = '';
        $requiredNew = '';

        switch ($form->specFinal[F_NAME]) {
            case 'form':
                $formName = $this->store->getVar(F_NAME, STORE_RECORD);
                break;
            case 'formElement':
                if (false !== ($formId = $this->store->getVar(FE_FORM_ID, STORE_SIP . STORE_RECORD))) {
                    $row = $this->databaseManager->getQfqDb()->sql("SELECT `f`.`name` FROM `Form` AS f WHERE `id`=" . $formId, ROW_EXPECT_1);
                    $rowRequiredNew = $this->databaseManager->getQfqDb()->sql("SELECT `f`.`requiredParameterNew` FROM `Form` AS f WHERE `id`=" . $formId, ROW_EXPECT_1);
                    $requiredNew = current($rowRequiredNew);
                    $formName = current($row);
                }
                break;
            default:
                return '';
        }

        if ($formName === false) {
            $buttonSpec['tooltip'] = "Form not 'form' or 'formElement'";
            $buttonSpec['disabled'] = 'true';
        } else {
            if ($requiredNew === '') {
                $requiredNew = $this->store->getVar(F_REQUIRED_PARAMETER_NEW, STORE_RECORD . STORE_EMPTY);
            }
            if (trim($requiredNew) !== '') {
                $buttonSpec['tooltip'] = "Form has 'required new' parameters and therefore cannot be previewed.";
                $buttonSpec['disabled'] = 'true';
            } else {
                $queryStringArray = [
                    'id' => $this->store->getVar(SYSTEM_EDIT_FORM_PAGE, STORE_SYSTEM),
                    'form' => $formName,
                    'r' => 0,
                ];
                $queryString = Support::arrayToQueryString($queryStringArray);
                $sip = $this->store->getSipInstance();
                $buttonSpec['url'] = $sip->queryStringToSip($queryString);

                $buttonSpec['tooltip'] = "View current form with r=0" . PHP_EOL . PHP_EOL . OnArray::toString($queryStringArray, ' = ', PHP_EOL, "'");
            }
        }

        $buttonSpec['buttonHtmlId'] = 'form-view-' . $form->specFinal[F_ID];
        $buttonSpec['icon'] = GLYPH_ICON_VIEW;
        $buttonSpec['class'] = "btn btn-default navbar-btn";
        return $this->buttonRenderer->renderElement($buttonSpec);
    }

    /**
     * Creates a Checkbox, which toggles 'hide'/'unhide' via JS, on all elements with class= CLASS_FORM_ELEMENT_EDIT.
     *
     * @return string - the rendered Checkbox
     */
    private function renderEditFormElementButton() {
        // EditFormElement Icons
        // TODO remove jquery dependency?
        $js = '$(".' . CLASS_FORM_ELEMENT_EDIT . '").toggleClass("hidden")';
        $element =  "<input type='checkbox' onchange='" . $js . "'>"
                    . Support::wrapTag("<span title='Toggle: Edit form element icons' class='"
                        . GLYPH_ICON . ' ' . GLYPH_ICON_TASKS . "'>", '');
        $element = Support::wrapTag('<label class="btn btn-default navbar-btn">', $element);

        return Support::wrapTag('<div class="btn-group" data-toggle="buttons">', $element);
    }

    /**
     * Creates a button to open 'CopyForm' with the current form as source.
     *
     * @return string - the rendered button
     * @throws \CodeException
     * @throws \UserFormException
     */
    private function renderCopyFormButton(Form $form) {
        $buttonSpec = array();

        // Show copy icon only on form 'form' and only if there is a form loaded (id>0)
        if ($form->recordId == 0 || $form->specFinal[F_NAME] != FORM_NAME_FORM) {
            return '';
        }
        // current loaded form.
        $formId = $this->store->getVar(COLUMN_ID, STORE_RECORD . STORE_ZERO);

        $queryStringArray = [
            'id' => $this->store->getVar(SYSTEM_EDIT_FORM_PAGE, STORE_SYSTEM),
            'form' => 'copyForm',
            'r' => 0,
            'idSrc' => $formId,
        ];
        $queryString = Support::arrayToQueryString($queryStringArray);
        $sip = $this->store->getSipInstance();
        $buttonSpec['url'] = $sip->queryStringToSip($queryString);
        $buttonSpec['icon'] = GLYPH_ICON_DUPLICATE;
        $buttonSpec['tooltip'] = "Duplicate form" . PHP_EOL . PHP_EOL . OnArray::toString($queryStringArray, ' = ', PHP_EOL, "'");
        $buttonSpec['class'] = 'btn btn-default navbar-btn';
        $buttonSpec['htmlButtonId'] = 'form-view-' . $form->specFinal[F_ID];

        if($formId == 0) $buttonSpec['disabled'] = true;

        return $this->buttonRenderer->renderElement($buttonSpec);
    }



}